#include "randomsearch.h"

RandomSearch::RandomSearch(Problem *p)
{
    myProblem = p;
    maxiters = 200;
    samples =20;
    iter = 0;
    lineSearchMethod = "none";//available values: armijo, none, golden,fibonacci
}

void    RandomSearch::setLineSearch(string m)
{
    lineSearchMethod = m;
}

string  RandomSearch::getLineSearch() const
{
    return lineSearchMethod;
}


double  RandomSearch::fl(double h)
{
    Data trialx = bestx;
    unsigned int i;
    Data g = myProblem->gradient(bestx);
    for(i=0;i<bestx.size();i++)
    {
        trialx[i]=trialx[i]-h * g[i];
    }
    return myProblem->statFunmin(trialx);
}

double RandomSearch::fib(int k)
{
    return (pow(1+sqrt(5),k+1)-pow(1-sqrt(5),k+1))/(pow(2,k+1)*sqrt(5));
}

void    RandomSearch::fibonacciSearch(double &a,double &b)
{

    int iteration = 2;
    while(true)
    {
          double phi = fib(iteration)/fib(iteration+1);
        double x1 = a+(1-phi)*(b-a);
        double x2 = a+phi * (b-a);
        double diff =fabs(x1-x2);
        if(diff<1e-6) break;

        if(fl(x1)<fl(x2))
        {
            b = x2;
        }
        else
        {
            a=x1;
        }
        iteration++;

    }
}
void    RandomSearch::armijoSerch(double &a)
{
    double f0 = myProblem->statFunmin(bestx);
    double beta =0.001;
    double t = 0.5;
    int iteration = 0;
    do
    {
            a = a * t;
            iteration++;
            if(iteration>=20) break;
    }while(fl(a)>f0-a*beta*myProblem->grms(bestx));

}

void RandomSearch::goldenSearch(double &a,double &b)
{
    double phi = (sqrt(5.0)-1.0)/2.0;
    int iteration = 1;
    while(true)
    {
        double x1 = a+(1-phi)*(b-a);
        double x2 = a+phi * (b-a);
        double diff =fabs(x1-x2);
        if(diff<1e-6) break;

        if(fl(x1)<fl(x2))
        {
            b = x2;
        }
        else
        {
            a=x1;
        }
        iteration++;

    }
}

void    RandomSearch::setMaxIters(int m)
{
    if(m>0) maxiters =m;
}

int     RandomSearch::getMaxIters() const
{
    return maxiters;
}

void    RandomSearch::setSamples(int s)
{
    if(s>0) samples=s;
}

int     RandomSearch::getSamples() const
{
    return samples;
}

void    RandomSearch::Solve()
{
    iter = 0;
    besty =1e+100;
    do
    {
        Step();
    }while(!Terminated());
}

bool    RandomSearch::Terminated()
{
    return iter>=maxiters;
}

void    RandomSearch::Step()
{
    ++iter;
    for(int i=0;i<samples;i++)
    {
        Data x= myProblem->getSample();
        double y = myProblem->statFunmin(x);
        if(y<besty)
        {
            besty = y;
            bestx =x;
            if(lineSearchMethod!="none")
            {
                double rate = 1.0;
                if(lineSearchMethod=="golden")
                {
                    double lambda = rate;
                    double a = lambda * 0.9;
                    double b = lambda * 1.1;
                    goldenSearch(a,b);
                    rate = a;
                }
                else
                if(lineSearchMethod == "fibonacci")
                {
                    double lambda = rate;
                    double a = lambda * 0.9;
                    double b = lambda * 1.1;
                    fibonacciSearch(a,b);
                    rate = a;
                }
                else
                if(lineSearchMethod == "armijo")
                {
                    double lambda  = 2;
                    armijoSerch(lambda);
                    rate = lambda;
                }
                Data g = myProblem->gradient(bestx);
                for(int j=0;j<bestx.size();j++)
                {
                    bestx[j]=bestx[j]-rate * g[j];
                }
                besty = myProblem->statFunmin(bestx);
            }
        }
    }
}

Data    RandomSearch::getBestx()
{
    return bestx;
}

double  RandomSearch::getBesty()
{
    return besty;
}

RandomSearch::~RandomSearch()
{
    //nothing here
}
